--------------------------------------------------------------------------------
-- CERN BE-CO-HT
--------------------------------------------------------------------------------
--
-- unit name:   vtu_diag
--
-- description: Frequency meter and counter
--
--------------------------------------------------------------------------------
-- Copyright CERN 2012-2019
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

--  Principle of operation:
--
--  This block counts the number of pulses on CLK_IN_I during a period.
--  At the end of the period, the value is saved and the counter reset.
--  The saved value is available on FREQ_O, which is synchronized with
--  CLK_SYS_I if G_SYNC_OUT is True.
--  The width of the counter is defined by G_COUNTER_BITS.
--
--  - If g_WITH_INTERNAL_TIMEBASE is True:
--    The period is defined by an internal counter that generates a pulse
--    every G_CLK_SYS_FREQ CLK_SYS_I ticks.
--
--  - If g_WITH_INTERNAL_TIMEBASE is False:
--    The period is defined by PPS_P1_I

library ieee;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gencores_pkg.all;

entity vtu_diag is
  generic (
    --  CLK_I frequency
    g_CLK_FREQ : natural := 62_500_000);
  port (
    clk_i        : in  std_logic;
    rst_n_i      : in  std_logic;

    --  Signal to be measured.
    sig_i        : in  std_logic;
    en_i         : in  std_logic;

    --  Select the window to measure the frequency.
    --  0: 500us/2kHz
    --  1: 50ms/20Hz
    --  2: 5s/0.2Hz
    --  3: 20s/0.05Hz
    --  A change to this value is taken into account at the end of the
    --  previous value.
    windows_i    : in std_logic_vector(1 downto 0);

    --  Last frequency measured.
    freq_o       : out std_logic_vector(31 downto 0);
    --  Counter incremented when FREQ_O is updated.
    generation_o : out std_logic_vector(15 downto 0);

    --  Count number of pulses.  Valid only if SIG_I pulses are longer than
    --  CLK_I period.
    counter_o    : out std_logic_vector(31 downto 0));
end vtu_diag;

architecture arch of vtu_diag is

  signal gate_pulse, gate_pulse_synced : std_logic := '0';

  signal cntr_last : unsigned(31 downto 0) := (others => '0');
  signal cntr_gate : unsigned(31 downto 0) := (others => '0');
  signal cntr_meas : unsigned(31 downto 0) := (others => '0');

  --  Result of the freq measure.
  signal freq_reg : std_logic_vector(31 downto 0) := (others => '0');

  signal generation : unsigned (15 downto 0);

  signal counter    : unsigned(31 downto 0);

  signal sig_pulse : std_logic;

  signal freq_wr : std_logic;
begin
  assert g_CLK_FREQ mod 2000 = 0
    report "inaccurate results due to clk frequency"
    severity failure;

  --  Mesure number of SIG_I pulses during the window.
  p_freq_counter : process (sig_i)
  begin
    if rising_edge(sig_i) then
      if gate_pulse_synced = '1' then
        freq_reg  <= std_logic_vector(cntr_meas);
        cntr_meas <= (others => '0');
      else
        if en_i = '1' then
          cntr_meas <= cntr_meas + 1;
        end if;
      end if;
    end if;
  end process p_freq_counter;

  --  Generate the end-of-window pulse at the end of the measure window.
  p_gate_counter : process(clk_i)
  begin
    if rising_edge(clk_i) then
      if cntr_gate = cntr_last then
        cntr_gate  <= (others => '0');
        gate_pulse <= '1';
        case windows_i is
          when "00" =>
            cntr_last <= to_unsigned (g_CLK_FREQ / 2000, 32) - 1;
          when "01" =>
            cntr_last <= to_unsigned (g_CLK_FREQ / 20, 32) - 1;
          when "10" =>
            cntr_last <= to_unsigned (g_CLK_FREQ * 5, 32) - 1;
          when "11" =>
            cntr_last <= to_unsigned (g_CLK_FREQ * 20, 32) - 1;
          when others =>
            null;
        end case;
      else
        cntr_gate  <= cntr_gate + 1;
        gate_pulse <= '0';
      end if;
    end if;
  end process;

  --  Synchronizer for the end-of-window pulse.
  U_Sync_Gate : entity work.gc_pulse_synchronizer2
    port map (
      clk_in_i  => clk_i,
      rst_in_n_i => rst_n_i,
      clk_out_i => sig_i,
      rst_out_n_i => '1',
      d_ready_o => open,
      d_ack_p_o => open,
      d_p_i     => gate_pulse,
      q_p_o     => gate_pulse_synced);

  inst_sig_sync : entity work.gc_sync_edge
    port map (
      clk_i => clk_i,
      rst_n_a_i => rst_n_i,
      data_i => sig_i,
      synced_o => open,
      pulse_o => sig_pulse);

  cmp_gc_sync_word_wr : entity work.gc_sync_word_wr
    generic map (
      g_AUTO_WR => TRUE,
      g_WIDTH   => 32)
    port map (
      clk_in_i    => sig_i,
      rst_in_n_i  => '1',
      clk_out_i   => clk_i,
      rst_out_n_i => rst_n_i,
      data_i      => freq_reg,
      data_o      => freq_o,
      wr_o        => freq_wr);

  process (clk_i)
  begin
    if rising_edge (clk_i) then
      if rst_n_i = '0' then
        generation <= (others => '0');
        counter <= (others => '0');
      else
        if freq_wr = '1' then
          generation <= generation + 1;
        end if;
        if sig_pulse = '1' then
          counter <= counter + 1;
        end if;
      end if;
    end if;
  end process;

  counter_o <= std_logic_vector (counter);
  generation_o <= std_logic_vector (generation);
end arch;
