--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- WR2RF_VME core
-- https://ohwr.org/project/vme-rf-wr-bobr
--------------------------------------------------------------------------------
--
-- unit name:   wr2rf_core
--
-- description:
--
--------------------------------------------------------------------------------
-- Copyright (c) 2019 CERN (home.cern)
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;

entity wr2rf_core is
    port (
        clk_i   : in  std_logic;
        rst_n_i : in  std_logic;

        --  Top-level wishbone port (from VME)
        wb_i               : in    t_wishbone_slave_in;
        wb_o               : out   t_wishbone_slave_out;

        --  Wishbone port for WR core
        wrc_wb_i           : in    t_wishbone_master_in;
        wrc_wb_o           : out   t_wishbone_master_out;

        --  Wishbone port for the VTUs
        rf1_vtus_wb_i          : in    t_wishbone_master_in;
        rf1_vtus_wb_o          : out   t_wishbone_master_out;

        rf2_vtus_wb_i          : in    t_wishbone_master_in;
        rf2_vtus_wb_o          : out   t_wishbone_master_out;

        mmcm_locked_i : in std_logic;
        clk_sel_o     : out std_logic;
        clk_sys_reset_o : out std_logic;

        tmg_clk_term_en_o : out std_logic_vector(2 downto 1);
        tmg_io_term_en_o  : out std_logic_vector(4 downto 1);
        tmg_clk_oe_o  : out std_logic_vector(1 downto 0);
        tmg_io_dir_o  : out std_logic;

        dds_cs_n_o         : out std_logic;
        rf1_iqdac_cs_n_o   : out std_logic;
        rf2_iqdac_cs_n_o   : out std_logic;
        rf_sclk_o          : out std_logic;
        rf_sdi_o           : out std_logic;
        rf_sdo_i           : in  std_logic;
    
        pll_main_cs_n_o : out std_logic;
        pll_main_sclk_o : out std_logic;
        pll_main_sdi_o  : out std_logic;
        pll_main_sdo_i  : in  std_logic;

        --  RF 1
        rf1_mux_sel_o        : out std_logic_vector(1 downto 0);
        rf1_mixer_en_o       : out std_logic;
        rf1_t1_delay_latch_o : out std_logic;
        rf1_t1_delay_oen_o   : out std_logic;
        rf1_t1_mux_sel_o     : out std_logic;
        rf1_t2_delay_latch_o : out std_logic;
        rf1_t2_delay_oen_o   : out std_logic;
        rf1_t2_mux_sel_o     : out std_logic;
    
        --  RF 2
        rf2_mux_sel_o        : out std_logic_vector(1 downto 0);
        rf2_mixer_en_o       : out std_logic;
        rf2_t1_delay_latch_o : out std_logic;
        rf2_t1_delay_oen_o   : out std_logic;
        rf2_t1_mux_sel_o     : out std_logic;
        rf2_t2_delay_latch_o : out std_logic;
        rf2_t2_delay_oen_o   : out std_logic;
        rf2_t2_mux_sel_o     : out std_logic;
    
        rf_delay_o           : out std_logic_vector(9 downto 0)
    );
end;

architecture arch of wr2rf_core is
    signal pll_spi_in : t_wishbone_master_in;
    signal pll_spi_out : t_wishbone_master_out;

    signal rf_spi_in : t_wishbone_master_in;
    signal rf_spi_out : t_wishbone_master_out;

    signal rf_in : t_wishbone_master_in;
    signal rf_out : t_wishbone_master_out;

    signal wrpc_wb16_in : t_wishbone_master_in;
    signal wrpc_wb16_out : t_wishbone_master_out;

    signal rf1_t1_delay_data, rf1_t2_delay_data, rf2_t1_delay_data, rf2_t2_delay_data : std_logic_vector (15 downto 0);
    signal rf1_t1_delay_wr, rf1_t2_delay_wr, rf2_t1_delay_wr, rf2_t2_delay_wr : std_logic;
begin
    inst_regs: entity work.wr2rf_vme_regs
      port map (
        rst_n_i => rst_n_i,
        clk_i => clk_i,
        wb_i => wb_i,
        wb_o => wb_o,
        ctrl_rf1_vtus_i => rf1_vtus_wb_i,
        ctrl_rf1_vtus_o => rf1_vtus_wb_o,
        ctrl_rf2_vtus_i => rf2_vtus_wb_i,
        ctrl_rf2_vtus_o => rf2_vtus_wb_o,
        ctrl_reg1_o => open,
        ctrl_reg2_o => open,
        init_clock_status_mmcm_locked_i => mmcm_locked_i,
        init_clock_ctrl_clk_sel_o => clk_sel_o,
        init_clock_ctrl_mmcm_reset_o => clk_sys_reset_o,
        init_tmg_io_term_o    => tmg_io_term_en_o,
        init_tmg_clk_term_o   => tmg_clk_term_en_o,
        init_tmg_clk_oe_o    => tmg_clk_oe_o,
        init_tmg_io_dir_o    => tmg_io_dir_o,

        init_pll_spi_i => pll_spi_in,
        init_pll_spi_o => pll_spi_out,

        init_rf_spi_i => rf_spi_in,
        init_rf_spi_o => rf_spi_out,

        init_rf_i => rf_in,
        init_rf_o => rf_out,

        init_wrpc_i => wrpc_wb16_in,
        init_wrpc_o => wrpc_wb16_out
      );

    inst_pll_spi: entity work.xwb_spi
      generic map (
        g_interface_mode => CLASSIC,
        g_address_granularity => BYTE,
        g_divider_len => open,
        g_max_char_len => open,
        g_num_slaves => 3
      )
      port map (
        clk_sys_i => clk_i,
        rst_n_i => rst_n_i,
        slave_i => pll_spi_out,
        slave_o => pll_spi_in,
        desc_o => open,
        int_o => open,  --  TODO: interrupt
        pad_cs_o (0) => dds_cs_n_o,
        pad_cs_o (1) => rf1_iqdac_cs_n_o,
        pad_cs_o (2) => rf2_iqdac_cs_n_o,
        pad_sclk_o => rf_sclk_o,
        pad_mosi_o => rf_sdi_o,
        pad_miso_i => rf_sdo_i
      );

    inst_rf_spi: entity work.xwb_spi
      generic map (
        g_interface_mode => CLASSIC,
        g_address_granularity => BYTE,
        g_divider_len => open,
        g_max_char_len => open,
        g_num_slaves => 1
      )
      port map (
        clk_sys_i => clk_i,
        rst_n_i => rst_n_i,
        slave_i => rf_spi_out,
        slave_o => rf_spi_in,
        desc_o => open,
        int_o => open,  --  TODO: interrupt
        pad_cs_o (0) => pll_main_cs_n_o,
        pad_sclk_o => pll_main_sclk_o,
        pad_mosi_o => pll_main_sdi_o,
        pad_miso_i => pll_main_sdo_i
      );

    my_inst: entity work.wr2rf_init_rf_regs
      port map (
        rst_n_i => rst_n_i,
        clk_i => clk_i,
        wb_i => rf_out,
        wb_o => rf_in,
        rf1_common_mixer_en_o => rf1_mixer_en_o,
        rf1_common_mux_sel_o => rf1_mux_sel_o,
        rf1_ch_0_csr_delay_oen_o => rf1_t1_delay_oen_o,
        rf1_ch_0_csr_mux_sel_o => rf1_t1_mux_sel_o,
        rf1_ch_0_delay_o => rf1_t1_delay_data,
        rf1_ch_0_delay_wr_o => rf1_t1_delay_wr,
        rf1_ch_1_csr_delay_oen_o => rf1_t2_delay_oen_o,
        rf1_ch_1_csr_mux_sel_o => rf1_t2_mux_sel_o,
        rf1_ch_1_delay_o => rf1_t2_delay_data,
        rf1_ch_1_delay_wr_o => rf1_t2_delay_wr,
        rf2_common_mixer_en_o => rf2_mixer_en_o,
        rf2_common_mux_sel_o => rf2_mux_sel_o,
        rf2_ch_0_csr_delay_oen_o => rf2_t1_delay_oen_o,
        rf2_ch_0_csr_mux_sel_o => rf2_t1_mux_sel_o,
        rf2_ch_0_delay_o => rf2_t1_delay_data,
        rf2_ch_0_delay_wr_o => rf2_t1_delay_wr,
        rf2_ch_1_csr_delay_oen_o => rf2_t2_delay_oen_o,
        rf2_ch_1_csr_mux_sel_o => rf2_t2_mux_sel_o,
        rf2_ch_1_delay_o => rf2_t2_delay_data,
        rf2_ch_1_delay_wr_o => rf2_t2_delay_wr
      );

    process(clk_i)
    begin
      if rising_edge(clk_i) then
        if rf1_t1_delay_wr = '1' then
          rf_delay_o <= rf1_t1_delay_data(9 downto 0);
        elsif rf1_t2_delay_wr = '1' then
          rf_delay_o <= rf1_t2_delay_data(9 downto 0);
        elsif rf1_t2_delay_wr = '1' then
          rf_delay_o <= rf2_t1_delay_data(9 downto 0);
        elsif rf1_t2_delay_wr = '1' then
          rf_delay_o <= rf2_t2_delay_data(9 downto 0);
        else
          rf_delay_o <= (others => '0');
        end if;
        rf1_t1_delay_latch_o <= rf1_t1_delay_wr;
        rf1_t2_delay_latch_o <= rf1_t2_delay_wr;
        rf2_t1_delay_latch_o <= rf2_t1_delay_wr;
        rf2_t2_delay_latch_o <= rf2_t2_delay_wr;
      end if;
    end process;

    inst_wb16x32: entity work.wb16_to_wb32
        port map (
          clk_i => clk_i,
          rst_n_i => rst_n_i,
          wb16_i => wrpc_wb16_out,
          wb16_o => wrpc_wb16_in,
          wb32_i => wrc_wb_i,
          wb32_o => wrc_wb_o
        );
end arch;
