library ieee;
use ieee.std_logic_1164.all;

use work.wishbone_pkg.all;
use work.trigunit_regs_consts.all;
use work.vtudiag_regs_consts.all;
use work.sim_wishbone.all;
use work.sim_wishbone16.all;

entity tb_vtu is
end tb_vtu;

architecture arch of tb_vtu is
    signal clk_sys : std_logic := '0';
    signal rst_sys : std_logic;

    signal clk_rf : std_logic := '0';
    signal clk_vtu : std_logic := '0';

    signal sync : std_logic := '0';
    signal trig_p : std_logic := '0';
    signal start, stop : std_logic := '0';

    signal wb_in : t_wishbone_slave_in;
    signal wb_out : t_wishbone_slave_out;

    function to_string(v : std_ulogic) return character is
    begin
      case v is
        when 'U' => return 'U';
        when '0' => return '0';
        when '1' => return '1';
        when 'X' => return 'X';
        when 'Z' => return 'Z';
        when 'L' => return 'L';
        when 'H' => return 'H';
        when 'W' => return 'W';
        when '-' => return '-';
      end case;
    end to_string;

    function to_string(v : std_logic_vector) return string is
      alias av : std_logic_vector(1 to v'length) is v;
      variable res : string(av'range);
    begin
      for i in av'range loop
        res (i) := to_string(av (i));
      end loop;
      return res;
    end to_string;

    --  Command to be sent to the VTU observer: number of expected pulses.
    type observer_type is record
      --  Number of expected pulses.
      count : natural;

      --  True if can go beyond the number of pulses (infinite window).
      infinite : boolean;

      --  If True, expect square waves.  Check half-period.
      square : boolean;
    end record;

    signal observer_cmd : observer_type;

    --  VTU observer state.
    type obs_state_type is (IDLE, WORKING, OBS_DONE, OBS_ERR);
    signal observer_state : obs_state_type := IDLE;
    signal observer_count : natural;
    signal observer_period : time;
begin
    --  System clock, 62.5 Mhz
    process
    begin
      clk_sys <= '0';
      wait for 8 ns;
      clk_sys <= '1';
      wait for 8 ns;
    end process;
    rst_sys <= '1', '0' after 16 ns;

    --  RF clock, ~200Mhz (5ns)
    --  VTU clock, ~25Mhz (40ns)
    process
      variable cnt : natural := 8;
    begin
      clk_rf <= not clk_rf;
      if cnt = 8 then
        clk_vtu <= not clk_vtu;
        cnt := 0;
      end if;
      cnt := cnt + 1;
      wait for 2.5 ns;
    end process;

    my_inst: entity work.vtu_blk
      port map (
        clk_sys_i => clk_sys,
        rst_sys_i => rst_sys,
        wb_i => wb_in,
        wb_o => wb_out,
        clk_vtu_i => clk_vtu,
        clk_rf_i => clk_rf,
        sync_i => sync,
        trig_o => trig_p,
        start_i => start,
        stop_i => stop
      );

    --  VTU observer.
    process (trig_p, observer_cmd)
      variable last_ts : time;
      variable this_period : time;
    begin
      if observer_cmd'event then
        --  New command
        case observer_state is
          when IDLE | OBS_DONE | OBS_ERR =>
            observer_state <= WORKING;
            observer_count <= 1;
            last_ts := now;
          when WORKING =>
            report "VTU observer command overrides the previous one" severity error;
        end case;
      end if;

      if trig_p'event and (observer_cmd.square or trig_p = '1') then
        report "Trigger:" & Natural'image(observer_count);
        case observer_state is
          when IDLE =>
            report "VTU observer: trigger before command" severity error;
          when WORKING | OBS_DONE =>
            this_period := now - last_ts;
            if observer_count = 2 then
              observer_period <= this_period;
            elsif observer_count > 2 then
               if this_period /= observer_period then
                 report "VTU observer: irregular period" severity error;
               end if;
            end if;
            last_ts := now;
            observer_count <= observer_count + 1;
            if observer_count = observer_cmd.count then
              observer_state <= OBS_DONE;
            elsif observer_count > observer_cmd.count and not observer_cmd.infinite then
              report "VTU observer: too many pulses" severity error;
            end if;
          when OBS_ERR =>
            null;
        end case;
      end if;
    end process;

    --  Test process.
    process
        --  Send a pulse on the start signal and wait until it is propagated to the VTU.
        procedure start_pulse is
        begin
          --  Start pulse.
          start <= '1';
          wait for 40 ns;
          start <= '0';

          --  Wait until the VTU get the start pulse
          for i in 1 to 7 loop
            wait until rising_edge(clk_vtu);
          end loop;
        end start_pulse;

        procedure sync_pulse is
        begin
          sync <= '1';
          wait for 10 ns;
          sync <= '0';

          --  Wait until the VTU get the sync pulse.
          for i in 1 to 3 loop
            wait until rising_edge(clk_vtu);
          end loop;
        end sync_pulse;

        procedure read_status (s : inout std_logic_vector(15 downto 0)) is
        begin
          read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_STATUS, s);
          report "status = " & to_string(s)
            & ", missvalid=" & to_string (s (TRIGUNIT_REGS_STATUS_MISSVALID_OFFSET))
            & ", missready=" & to_string (s (TRIGUNIT_REGS_STATUS_MISSREADY_OFFSET))
            & ", startready=" & to_string (s (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET))
            & ", running=" & to_string (s (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET));
          -- assert s (TRIGUNIT_REGS_STATUS_WRONGBVALUE_OFFSET) = '0' severity error;
          -- assert s (TRIGUNIT_REGS_STATUS_WRONGHTVALUE_OFFSET) = '0' severity error;
          -- assert s (TRIGUNIT_REGS_STATUS_WRONGWVALUE_OFFSET) = '0' severity error;
        end read_status;

        variable val : std_logic_vector(15 downto 0);
    begin
        sync <= '0';
        start <= '0';

        wait until rst_sys = '0';

        --  Program the diag (500us)
        write16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_TRIGDIAG + ADDR_VTUDIAG_REGS_CONTROL, x"0000");

        --  Program the vtu
        --  Delay between start and the first pulse.
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_BVALUEOFFLINE, x"0000_0000_0000_0020");
        --  Delay between the pulses.
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_HTVALUEOFFLINE, x"0000_0000_0000_0008");
        --  Number of pulses.
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEOFFLINE, x"0000_0000_0000_0040");

        --  Windowed mode, enable.
        write16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_CONFIGOFFLINE, x"0021");

        --  Remove reset.
        write16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_CONTROL, x"0000");

        --  Check status.
        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_MISSVALID_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_MISSREADY_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET) = '0' severity error;
        --  Set to one after reset.
        assert val (TRIGUNIT_REGS_STATUS_WRONGBVALUE_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_WRONGHTVALUE_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_WRONGWVALUE_OFFSET) = '1' severity error;

        --  Check online config is 0.
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_CONFIGONLINE, val);
        report "config online = " & to_string(val);
        assert val = x"0000" severity error;

        --  Expect 64 syncs.
        observer_cmd <= (count => 64, infinite => False, square => False);

        --  Start pulse.
        start_pulse;

        --  Check status.
        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_MISSVALID_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_MISSREADY_OFFSET) = '0' severity error;
        --  Should be 0 if close enough to the start pulse.
        assert val (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_WRONGBVALUE_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_WRONGHTVALUE_OFFSET) = '0' severity error;
        assert val (TRIGUNIT_REGS_STATUS_WRONGWVALUE_OFFSET) = '0' severity error;

        --  Sync pulse.
        sync_pulse;

        --  Check ...
        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET) = '1' severity error;

        --  Check online config.
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_CONFIGONLINE, val);
        report "config online = " & to_string(val);
        assert val = x"0020" severity error;

        --  Check online registers
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEONLINE + 6, val);
        assert val = x"0040" severity error;
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEONLINE + 4, val);
        assert val = x"0000" severity error;
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEONLINE + 2, val);
        assert val = x"0000" severity error;
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEONLINE + 0, val);
        assert val = x"0000" severity error;
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_HTVALUEONLINE + 6, val);
        assert val = x"0008" severity error;
        read16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_BVALUEONLINE + 6, val);
        assert val = x"0020" severity error;

        --  Wait until end of generation.
        wait on observer_state;
        --  Check observer status.
        assert observer_state = OBS_DONE severity error;
        assert observer_period = 8 * 5 ns;
        --  Check VTU status.
        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET) = '0' severity error;

        -------------------------------------------------------------------------------------------------

        --  Test 2: check that missing valid parameters is reported.
        --  (A cycle has been finished, the start signal has to be ignored because
        --   parameters aren't set.)
        report "test 2 - start";
        start_pulse;

        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_MISSVALID_OFFSET) = '1' severity error;

        -------------------------------------------------------------------------------------------------

        --  Test 3: Program the vtu
        --  Delay between start and the first pulse.
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_BVALUEOFFLINE, x"0000_0000_0000_0010");
        --  Delay between the pulses.
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_HTVALUEOFFLINE, x"0000_0000_0000_0011");
        --  Number of pulses (unused).
        write64be_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_WVALUEOFFLINE, x"0000_0000_0000_0014");

        --  Lowfreq generation, enable.
        write16_pl (clk_sys, wb_in, wb_out, ADDR_TRIGUNIT_REGS_CONFIGOFFLINE, x"0041");

        --  Expect 64 syncs.
        observer_cmd <= (count => 20, infinite => true, square => true);

        report "test 3 - start";
        start_pulse;

        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_MISSVALID_OFFSET) = '0' severity error;

        --  Sync pulse.
        sync_pulse;

        --  Check ...
        read_status(val);
        assert val (TRIGUNIT_REGS_STATUS_STARTREADY_OFFSET) = '1' severity error;
        assert val (TRIGUNIT_REGS_STATUS_RUNNING_OFFSET) = '1' severity error;

        --  Wait until end of generation.
        wait on observer_state;
        --  Check observer status.
        assert observer_state = OBS_DONE severity error;
        assert observer_period = 17 * 5 ns;

        --  Stop counter.  Synchronize to
        wait until rising_edge(clk_vtu);
        stop <= '1';
        wait for 40 ns;
        stop <= '0';

        --  Wait a little bit (2 full periods)
        wait for 2 * 2 * observer_period;

        --  We should have stopped after 20 periods.  There are some extra periods because of synchronizers.
        assert observer_count = 23 severity error;

        report "end of tests";
        wait;
    end process;
end arch;
